/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */
 

//  xxxxxxxxxxxxxxx     wm15 global variable declarations  xxxxxxxxxxxxxxxx


// These variables below can be directly modified by the user on the options form: 

var item_cnt= 0;       // The number of stimuli in the task

var stype = '1';      //  Stimulus type: "0" => Alphanumeric characters; "1"  => Images (selected by the user)

var alfa_type = '0';  // Type of Alphanumeric characters => "0" = Digits; "1" = Letters-lowercase; "2" = Letters-UPPERCASE; "3" = All digits and letters

var font_family = 'Georgia, serif';  // user selected Font Family. canvas ctx.fontFamily (see drawText function)

var raster = 3;       // Raster size for each stimulus 
 
var model= true;   // the correctly composed image (model) will be displayed on the screen: true == Yes, false == No.   
	
var two = false;    // Number of canvases (Drag and drop region) in the task: Two canvas == true; One canvas == false.
 
var rotation = true;  // Parts of the stimulus will be rotated (true) or The stimulus parts are not rotated (false)

var ang = 180;        // rotation angle increament (degrees). Rotation step (degrees) for parts (puzzles)

var zoom_out = false;   // Zoom out focused image part is on (true) or off (false). It is only enabled when rotation is used.

var is_border = false;  // Separate the puzzles with the border (true) or not (false)
 
var demo = false;     // if (demo == true) => Solve the puzzle task 

var timer = 600;     // Task time in seconds if clock = 1000ms (default). 

var timer_enabled = false;   // if true => timer is enabled. if false timer is disabled. Exactly: the timer is always on

var is_footer= true;   // Footer is visible (true) or no (false)

var is_back = true;    // if (is_back==true) => user can go back to previous task/item. Footer must be visible  

var is_forward = true;  // if (is_forward==true) => user can escape (skip) items. Next task/item button vill be visible.  Footer must be visible

var prg_type = "0";   //  type of information in Progress Bar ("0" => Items, "1" => Time; "2" => Nothing) 

var back_clr= '#e5e5e5';  // background window color during task. Don't use white color (s_canvas + r_canvas have white colors) 

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 


// Other global constants and variables:

const clock = 1000;  // basic timer resolution in ms. Default = 1s (1000ms)

var timer_id = null;   // timer_id is the identifier of the repeated action you want to cancel. This ID is returned from setInterval().

var counter = 0;      // timer counter. if (counter >= time) Timer will call theEnd()   

var is_task = false;    // task in progress or not (for timer)

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var item = 0;   // current item in task     

var stim = [];   // stimulus content displayed to the user (alphanumeric characters) or image.titles

// These 2 arrays (s_parts and r_parts) are for display the current status on the screen when navigating through an item.

var s_parts= [];  // start canvas =>  [item] x [rect position] x [ 0=correct position, 1=rotation angle in degrees ] 

var r_parts= [];   // result canvas => [item] x [rect position] x [ 0=correct position, 1=rotation angle in degrees ]. This array is used to calculate the total score. 

var shown = [];   // for init each item: if true. rect init is done for this item. see loadImage() 

var moves = []; // For each item: number of moves => see drop () event

var angles = [];  // set of all rotation angles for parts (puzzles)

var rot_cnt= [];  // total number of rotation done by user for each item (only if rotation is enabled). See doubleClick() event

var go_forward = false; // go to next item when current item is done (flag for timer).

var delay_cnt;     //  it gives about 2s of time delay when item is done. works in pair with go_forward. see appTimer()


//  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


// application init [create/start Timer, check storage, show start panel..] 
function init() 
{  
  if (timer_id)  {   
    window.clearInterval(timer_id);  
  }  
  timer_id= setInterval(function () {appTimer()},clock);  
    
  if (isStorage() && storageAvailable('localStorage'))  
  {
    document.getElementById("store_group").disabled= false;
  }
  else 
  {
    document.getElementById('store_msg').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline).";
    document.getElementById('store_group').disabled= true;  
    document.getElementById('clr_store').disabled= true;     // IE11 bug  => childrens are not auto disabled
    document.getElementById('store').disabled= true;          // IE11 bug => childrens are not auto disabled
  }    
	
	if ( !readerExists() ) {
	  document.getElementById('stype').disabled= true;
		document.getElementById('no_reader').style.display= 'block';		
	}	
		  
	setIns();
	
	loadFont();
	
	showStart(); 
 
}


// application timer handler function
function appTimer()  
{ 
  if (!is_task) {
		return;  // nop, idle, do nothing        
  }	   
	if (timer_enabled) 
	{
		counter++;  
		if (counter >= timer)  {     // task time is limited
			theEnd();
			return;
		}	
		if (prg_type == "1")   {   // "1" => Time. display time on Progress Bar  
			let rest= timer-counter;   // this is remaining time 
			let percent= Math.round(100*rest/timer);    
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('time_left').textContent= stringTime(rest);    
		}
	}
  if (go_forward)  
  {
    delay_cnt++;
    if (delay_cnt >= 2) { 
      nextItem();  //go_forward= false;      
    }
  }	
}	


// return integer into string hh:mm:ss time format
// input par: total == integer (time in seconds)
function stringTime(total) 
{
  var hh = Math.floor(total/(60*60));         
  var divisor_for_minutes = total % (60 * 60);
  var mm = Math.floor(divisor_for_minutes / 60);       
  var divisor_for_seconds = divisor_for_minutes % 60;
  var ss = Math.ceil(divisor_for_seconds);    
  // These lines ensure you always have two-digits. This formats your string to HH:MM:SS
  if (hh < 10)  {hh = "0"+hh;}     
  if (mm < 10) {mm = "0"+mm;}
  if (ss < 10) {ss = "0"+ss;} 
  return hh + ":" + mm + ":" + ss;
} 


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 and Edge offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// checks if the client's web browser supports this: FileReader object, Blob object, FileReader.readAsDataURL() 
function readerExists() 
{
	try
	{		
		// create Blob from a typed array and strings: https://javascript.info/blob
		let hello = new Uint8Array([72, 101, 108, 108, 111]); // "Hello" in binary form
		let blob = new Blob([hello, ' ', 'world'], {type: 'text/plain'});		
		let reader = new FileReader();
		if (blob) {
			reader.readAsDataURL(blob);
			return true;
		}
	}	 
  catch(e) { 
    return false;
  }
	return false;
}


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{ 
	if (is_task) { 
		drawStim();
	}
}	


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement("option");
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex=1; // Georgia, serif
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[1];
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.body.style.backgroundColor= '#ffffff';  // restore white color
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  if ( document.getElementById('store_group').disabled === false && localStorage.getItem('wm15-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('wm15-v1.0')).length;                                                        
    document.getElementById("store_cnt").textContent=  'You have a total of ' + cnt  + ' wm15 results stored in your web browser.';
  }
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		
  document.getElementById('ins_model').style.display= model ? 'block' : 'none';
	document.getElementById('ins_two').style.display= two ? 'block' : 'none';
  
  if (rotation)  {
    document.getElementById('ins_rot').style.display= 'block';
    document.getElementById('ins_ang').textContent= ang;
  }
  else  {
    document.getElementById('ins_rot').style.display= 'none';
  }
 
  document.getElementById('ins_prg').setAttribute("style", "display:none");
	if ( is_footer && ( prg_type === '0' || document.getElementById('timer_enabled').checked ) )
  {   
    document.getElementById('ins_prg').setAttribute("style", "display:block"); 
    if (document.getElementById('timer_enabled').checked)  
    {		
      document.getElementById('ins_prg').textContent= 'The duration of this task is limited.';  // The duration of this task is always limited.
      if (prg_type === "0")  {
        document.getElementById('ins_prg').textContent+= ' The Progress bar will show how many items (images) are left until the end of the task.';
      }
      else if (prg_type === "1")  {
        document.getElementById('ins_prg').textContent+= ' The Progress bar will display the remaining time until the end of the task.';  
      }
      else  {
        document.getElementById('ins_prg').textContent+= ' But information about the remaining time will not be visible.';  
      }  
    }
    else  { 
      document.getElementById('ins_prg').textContent= 'The Progress bar will show how many items (images) are left until the end of the task.';
    }			
  }
  
  document.getElementById('ins_back').style.display = is_footer && is_back ? 'block' : 'none';
	document.getElementById('ins_forward').style.display = is_footer && is_forward ? 'block' : 'none';
	document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';	
	
}	

  
// refresh Options (update *_msg label) => give user short option HELP
function refOption(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 'item_cnt':
			itemsValid();
			break;
		
		case 'stype':
		  if (e.value === '1')  {
				msg.textContent= 'Images will be displayed (you must select image files)';
				document.getElementById('alfa_opt').style.display= 'none';        // document.getElementById("alfa_opt").disabled= true;
				document.getElementById('images_opt').style.display= 'block';   // 	document.getElementById("images_opt").disabled= false;
			}
			else  {
				msg.textContent= 'Random alphanumeric characters will be displayed. Tip: use small raster for Alphanumeric characters.'; 
			  document.getElementById('images_opt').style.display= 'none';  // 	document.getElementById("images_opt").disabled= true;
				document.getElementById('alfa_opt').style.display= 'block';    // document.getElementById("alfa_opt").disabled= false;
			}	
      itemsValid();     		
			break;
				
		case 'raster':  
			if (e.validity.valid)  {           
				msg.removeAttribute("class");  
        let tmp= parseInt(e.value, 10) * parseInt(e.value, 10);        
				msg.textContent= 'Number of parts= ' + tmp + '. If raster is higher, the stimulus will be broken down into a larger number of parts (puzzles) and vice versa.';
        //msg.textContent+= ' Lower value = task is easier. Higher value = the task is more difficult';    
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 					
			break; 		
				
    case 'model':
			msg.textContent= e.value === '0' ? 'A correctly composed image (model) will not be displayed on the screen.' : 'The correctly composed image (model) will be displayed on the screen.';
			break;	
		
		case 'two':
			 msg.textContent= e.value === '0' ? 'Only one canvas will be displayed. Using the drag and drop action, the pieces (puzzles) need to be rearranged into the correct positions within the same canvas.' : 'Two canvases will be displayed: the first with mixed pieces (puzzles) and the second blank canvas. It is required to move the parts from the first canvas to the correct position in the second canvas with the Drag and drop operation.' ; 		
			break;
		
    case 'rotation':   
			if (e.checked)	{
				msg.textContent= 'Parts of the stimulus will be rotated';
        document.getElementById('rot_opt').style.display= 'block';			
			}
			else	{
				msg.textContent= 'The stimulus parts are not rotated'; 
        document.getElementById('rot_opt').style.display= 'none';	
			}	   			
			break; 
		
		 case 'ang':
       let ind = e.selectedIndex || 0 ;
			 msg.textContent= document.getElementById('ang' + ind).getAttribute('data-help');
       break;		 
		    
     case 'zoom_out':
       msg.textContent= e.checked ? 'Decrease the apparent size of focused part of an image, to view it as if from a greater distance or from a wider perspective. It can be useful if the puzzles overlap during rotation.' : 'All rotating parts of the image are displayed in normal size. Spatial overlapping of parts of the image is possible.';
       break;
       
		case 'is_border':
		  msg.textContent= e.checked ? 'Rectangle grid will be displayed on canvas' : 'No grid'; 
			break;
				
    case 'timer':
    if (e.validity.valid)  {
      msg.textContent= stringTime(e.value); 
      msg.removeAttribute('class');
    }
    else  {
      msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
      msg.setAttribute('class', 'error');
    } 
      break;                
      
    case 'timer_enabled':   
      if (e.checked)  {
        msg.textContent= 'Yes';
        document.getElementById('timer').disabled= false;
      }
      else  {  
        msg.textContent= 'Timer is disabled';
        document.getElementById('timer').disabled= true; 
      }       
      break;
				
		case 'is_footer':
			if (e.checked)  {
				msg.textContent= 'Yes';
				document.getElementById('footer_opt').style.display= 'block';
			}
			else  {
				msg.textContent= 'No';
				document.getElementById('footer_opt').style.display= 'none';
			}
			break;
					
    case 'demo':    	
    case 'is_back':     		
			msg.textContent= e.checked ? 'Yes' : 'No';    
			break;
    
    case 'is_forward':
    	msg.textContent= e.checked ? 'Yes' : 'If the footer or the button for the next task are not visible, the user must successfully solve the puzzle to display the new image.';    
			break;
    
    case 'prg_type':                                                                                                                                            
      if (e.value === "0")  {
        msg.textContent= 'It will be displayed how many items are left to the end.';
      }
      else if (e.value === "1")  {
        msg.textContent= 'The remaining time until the end of the task will be displayed. Only when timer is enabled.';
      }
      else  {
        msg.textContent= ''; //'Nothing';
      }
      break;    
		
		case 'back_clr': 
			if (validColor(e.value))   {           
				msg.removeAttribute('class');       
				msg.innerHTML= "Background color during task.<b>Warning: Don't use white color.</b>";  
			}
			else  {
				msg.textContent= 'The background color is invalid.';
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 
			break; 
		 
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// check if items count is valid or not
function itemsValid()  
{
  let e= document.getElementById('item_cnt'), msg= document.getElementById('item_cnt_msg'), s = 'Values within this range are allowed: ', s2= ' with step= '; 
	let stim_type= document.getElementById('stype').value;  // stimulus type
	let images= document.getElementById('preview').getElementsByTagName('img');  
	let img_cnt = (images && images.length) ? images.length : 0;
	if ( ( e.validity.valid === false ) ||
       (( stim_type === '1' && parseInt(e.value, 10) > img_cnt ) ) )  {
		msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
		if ( stim_type === '1') {
			msg.textContent+= '. For Stimulus type Images: Item count can not be greater then images count. Images count=' + img_cnt; 
    }		
		msg.setAttribute('class', 'error');    // msg.className = 'error';
		return;		
	}	
	msg.removeAttribute("class");       
	msg.textContent= (stim_type === '1') ? 'The number of stimuli (images) in the task' : 'The number of stimuli (random alphanumeric characters) in the task';   
}  


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  (dirty===true) ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// check for validity.valid object in all input type="number" + other internal app checks.see: setPar()
function optionsValid()  
{     
  let res= document.getElementById('item_cnt').validity.valid && document.getElementById('raster').validity.valid && document.getElementById('timer').validity.valid;	
	if (document.getElementById('stype').value == "1" && document.getElementById('preview').getElementsByTagName('img').length === 0) {	
	  res= false;
	}
  return res;              
}


// start task (research) 
function startTask(e)   
{     
  if (stype === '1' && document.getElementById('preview').getElementsByTagName('img').length === 0) {		 
		document.getElementById('is_valid').textContent= 'Unable to start task because you did not select image files to use as stimuli. Go to "Options" => "Select image files" or choose Stimulus type: Alphanumeric characters'; 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold');
    return;		
	}  
	
	setPar();
 
	setStim();
	
	//document.body.style.backgroundColor= '#e5e5e5';
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
  
	document.getElementById('task_section').style.display= 'block'; 
	document.getElementById('footer').style.display= is_footer ? 'block' : 'none';  	
	document.getElementById('progress').style.display= 'block'; 
  document.getElementById('progress').setAttribute('style', 'width:100%');
	
  let s_canvas= document.getElementById('s_canvas'), 
    r_canvas= document.getElementById('r_canvas'), 
    m_canvas= document.getElementById('m_canvas'),
    my_img= document.getElementById('img_cont');
  
  my_img.style.display= 'block';
	s_canvas.style.display= two ? 'block' : 'none';
	r_canvas.style.display= 'block';
  
	my_img.style.position='absolute';	
	s_canvas.style.position='absolute';
	r_canvas.style.position='absolute';
	m_canvas.style.position='absolute';
	
  if (two)  {
     s_canvas.title= 'Use the drag and drop action to move parts of the image to the white part of the screen'; // can be hidden by m_canvas
     r_canvas.title= 'Move parts of the image here'; // 'Use the drag and drop action to move parts of the image here'; // can be hidden by m_canvas
     m_canvas.title= rotation ? 'Use the drag and drop action to move parts of the image to the white part of the screen. Use a double-click to rotate clockwise' : 'Use the drag and drop action to move parts of the image to the white part of the screen' ; 
  }
  else  {
    s_canvas.title= ''; 
    r_canvas.title= 'Use the drag and drop action to move part of the image to the correct position';
    m_canvas.title= rotation ? 'Use the drag and drop action to move part of the image to the correct position. Use a double-click to rotate clockwise' : 'Use the drag and drop action to move part of the image to the correct position'; 
  }
  my_img.title= 'An exactly rearranged puzzle should end up looking like this'; 
  
	if (document.getElementById("prg_type").value === "2") {
	  document.getElementById("progress_holder").style.display= "none"; 
	}	
	else  
	{
		document.getElementById('progress_holder').style.display= 'block';     
		document.getElementById('progress').setAttribute('style', 'width:100%');   		
		if (document.getElementById('prg_type').value === '0') {
		  document.getElementById('progress').className='progress-bar progress-bar-info progress-bar-striped';     // items/total
		}
		else if (document.getElementById('prg_type').value === '1')  {
		  document.getElementById('progress').className= 'progress-bar progress-bar-danger progress-bar-striped';  // remaining time   
		} 	
		document.getElementById('progress').setAttribute('style', 'width:0');  // reset progress bar width
	}
					
	if (validColor(back_clr) ) 	{
		try  {
			document.body.style.backgroundColor= back_clr;      
		}
		catch(e) {
      document.body.style.backgroundColor= '#e5e5e5';
      back_clr= '#e5e5e5';			
		}				
	}	  	
	document.getElementById('time_left').textContent='';  // reset 
  
  item= 0;
  
	counter= 0;       // reset timer counter
	
  is_task= true;    // finally "activate" timer  	
		
	drawStim();     // show (draw) first stimulus
      
}


// set internal application research parameters (global vars)
function setPar()  
{  
  stim= []; 
  
  shown=[];
  s_parts= [];  // start canvas =>  [item] x [rect position] x [ 0=correct position, 1=rotation angle in degrees ]   
  r_parts= [];  // result canvas => [item] x [rect position] x [ 0=correct position, 1=rotation angle in degrees ]
  moves= [];  
  rot_cnt = [];
	  	
	if (stype === '1' && item_cnt > document.getElementById('preview').getElementsByTagName('img').length) {		 
		item_cnt= document.getElementById('preview').getElementsByTagName('img').length;
		document.getElementById("is_valid").textContent= "The number of items was greater than the number of images. The number of items is automatically reduced to: " + item_cnt; 
    document.getElementById("is_valid").setAttribute("style", "text-align:center;color:#ff0000;font-weight:bold"); 
	}  
    
  for (let i= 0; i< item_cnt; i++)   
  { // i => item
    shown[i]= false;
    s_parts[i]= [];  
    r_parts[i]= [];	
    rot_cnt[i]= 0;		
    moves[i]= 0;  // number of moves for each item
    for (let j= 0; j< raster*raster; j++)  {     
      s_parts[i][j]= [-1,0];  // correct position, current angle = 0 (no rotation)
			r_parts[i][j]= [-1,0];  // correct position, current angle = 0 (no rotation)    	
    }
  }
		
	angles=[]; 
	if (rotation)  {  
		for (let i= 0; i < 360/ang; i++) {
			angles[i]= ang * i;
		}									
	}
  
}


// generates task stimuli 
function setStim()	
{
	if (stype === '0')  // alphanumeric characters are selected
	{  
		let s, previous;		
		for (let j= 0; j<item_cnt; j++)   
		{		
			do
			{	
			  s='';
				switch (alfa_type)
				{
					case '0':  // digits 								
							s= getRndInt(0,9).toString();								  
						break;			
					case '1':  // Letters-lowercase 
						s= lowerStrGen();			
						break;
					case '2':  // Letters-UPPERCASE 
						s= upperStrGen();			
						break;
					default:  // All digits and letters 
						s= allStrGen();
				} 
			} while (s === previous);
			stim[j]=s;
			previous=s;		
		}
	}
	else  // images are selected
	{
		let image;
		for (let j= 0; j<item_cnt; j++)   {
			image= document.getElementById('preview').getElementsByTagName('img')[j];
			stim[j]= image.title ? image.title : 'image ' + parseInt(j+1);  // images file names array => document.querySelector('input[type=file]').files[i].name
			//let is_title= image.getAttribute('data-title');
			//stim[j]= is_title ? is_title : 'item ' + parseInt(j+1);  // images file names array => document.querySelector('input[type=file]').files[i].name
		}					
	}
}


// generates a random string from lowercase letters
function lowerStrGen()
{ 
  let all = 'abcdefghijklmnopqrstuvwxyz';
  let tmp= all.length;  
  let txt= all.charAt(Math.floor(Math.random() * tmp));	
  return txt;
}


// generates a random string from uppercase letters 
function upperStrGen()
{ 
  let all = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  let tmp= all.length;
  let txt = all.charAt(Math.floor(Math.random() * tmp));	
  return txt;
}


// generates a random string from all (uppercase and lowercase) letters and digits
function allStrGen()
{ 
  let all = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
  let tmp= all.length;
  let txt= all.charAt(Math.floor(Math.random() * tmp));	
  return txt;
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
  stype= document.getElementById('stype').value;  // must be updated before item_cnt
	
	if (document.getElementById('item_cnt').validity.valid)  {  
    if ( stype === '0' || (stype === '1' && parseInt(document.getElementById('item_cnt').value, 10) <= document.getElementById('preview').getElementsByTagName('img').length) ) {
			item_cnt= parseInt(document.getElementById('item_cnt').value, 10);			
		}
  } 
	 	      	
  alfa_type= document.getElementById('alfa_type').value;
		
	font_family= document.getElementById('font_family').value;
	
  if (document.getElementById('raster').validity.valid)  {   
    raster= parseInt(document.getElementById('raster').value, 10);   
	}

  model= document.getElementById('model').value === '1' ? true: false;
	
	two = document.getElementById('two').value === '1' ? true: false;
	
	rotation = document.getElementById('rotation').checked ? true : false;
	
  ang = parseInt(document.getElementById('ang').value, 10);
	
  zoom_out = (rotation && document.getElementById('zoom_out').checked) ? true: false;
  
	is_border = document.getElementById('is_border').checked ? true: false; 
	
	demo= document.getElementById('demo').checked ? true: false; 
 	
	if (document.getElementById("timer").validity.valid) timer= Math.round(parseInt(document.getElementById("timer").value, 10));  // task time (s)
  
	timer_enabled= (document.getElementById("timer_enabled").checked) ? true: false;  // timer is enabled or disabled
		
	is_footer= document.getElementById('is_footer').checked ? true: false;  // footer is visible or not
	
	is_back= document.getElementById('is_back').checked ? true: false;   // enable Go back (previous item)      
  
  is_forward= document.getElementById('is_forward').checked ? true: false; // enable forward (skip item)
	
	prg_type= document.getElementById('prg_type').value;   //  type of information displayed on Progress Bar
	
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
		     
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}


// get random integer from min to max range
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1)) + min;
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{  
  go_forward=false;                       
  document.body.style.backgroundColor= back_clr;
  if (item > 0)  {
    item--;  
    drawStim();     // draw again previous stimulus
  }         
}


// goto next item (stimulus)
function nextItem()  
{  
  go_forward=false;
  document.body.style.backgroundColor= back_clr;
	item++;			
		
	if (item < item_cnt)  {         		 
    drawStim();     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


// draw (repaint) stimulus. This is dispatch function (call other)
// call in startTask() + each body onresize() when is_task == true => call this function      
function drawStim() 
{    		
	// this function always implicity call loadImage when set document.getElementById('img_cont').src=.. 
	if (stype == '0')  {   // Random alphanumeric characters will be displayed    		
	  setPositions();
		drawText(stim[item]); 		
		let bounds= setTextBounds();
    canvasToImage(bounds);				
	}
	else	{		// Images will be displayed (user must select image files)
    document.getElementById('img_cont').src= document.getElementById('preview').getElementsByTagName('img')[item].src;   //document.getElementById('img_cont').crossOrigin= 'anonymous'; // If crossOrigin is an empty string (""), the anonymous mode is selected.		 		
	}		
 	
  if (is_footer)
	{  // footer is visible
		document.getElementById("prior").style.display= (item > 0 && is_back) ? 'inline': 'none';  // set footer navigation prior button (left arrow) 		
		if (is_forward)     
		{   // set footer next item button (right arrow) 
			let btn= document.getElementById("next");     
			if(item === item_cnt-1)  {       
        btn.style.backgroundImage=  "url('img/end.png')";  // update "Next item" image on "right arrow" button 
        btn.title= 'End of task';
      }
      else  {
        btn.style.backgroundImage= "url('img/forward.png')"; 
        btn.title= 'Next image';
      }
      btn.style.display= "inline";  
		}
		else {
			document.getElementById("next").style.display= "none";
		}						
		if (prg_type == '0' )  {    // update Progress bar on footer => show current item count and remaining item count on Progress Bar (default)                                                                                                                                                       
			let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
			document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
			document.getElementById('time_left').textContent= 'Item= ' + (item + 1) + ' Remaining= ' + (item_cnt - item-1);  
		} 
	}
}


// draw text (stimulus) on r_canvas 
function drawText(txt)  
{
	let r_canvas= document.getElementById('r_canvas');
  clear(r_canvas);
	let ctx= r_canvas.getContext('2d');
	let x= r_canvas.width/2;
	let y= r_canvas.height/2;	     
	ctx.fillStyle = "#000000";  // line color 	
	let dx= Math.floor(0.5 * r_canvas.width);  
	let dy= Math.floor(0.5 * r_canvas.height); 
	let rad= Math.floor(Math.min(dx,dy)/2);  // text fontSize	                                                                   
	ctx.font= rad.toString() + 'px ' + font_family; 
	ctx.textBaseline= 'bottom';  
  ctx.textBaseline= 'middle';    
	ctx.textAlign= 'center';	
	ctx.fillText(txt, x, y);
}	


// draw image from ../images directory into r_canvas object (html container)
function loadImage()  
{	    
	setPositions();
		
 	if (!shown[item])  
  {	   
    if (rotation)  {
      initRotatedParts();  
    }
    else  {
      initParts();   
    } 				
    shown[item]= true;
  }
  
	if (rotation)  {
		showRotatedParts();  
	}
	else  {
		showParts();
	}			
  	
	
  if (is_border) { 
		gridRect();	
  }	
}


// set image and canvas positions on screen (depending on screen resolution)
function setPositions()
{
  if (two === true) {
		if (model) {
			setPosModelTwo();
		}
		else  {
			setPosNoModelTwo();
		}
  }
	else {   // two == false => only r_canvas is visisble
		if (model) {
			setPosModelOne();
		}
		else  {
			setPosNoModelOne();
		}	
	}	
     
  document.getElementById('m_canvas').style.display= 'none';	// set moveable canvas (drag and drop) invisible
  	
	let ctx= r_canvas.getContext('2d');
  ctx.fillStyle = "#ffffff";    
  ctx.fillRect(0, 0, r_canvas.width, r_canvas.height);		
		
}	


// set absolute positions with pure js for: visible image (model) + 2 canvas (s_canvas + r_canvas)
function setPosModelTwo()
{  
	let win_w= window.innerWidth, win_h= window.innerHeight;  	     
		// tot_h == total y height => window height - footer height	
	let tot_h= win_h - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight              
  	
  let hor = win_w / tot_h  >= 1.5 ? true: false;  // hor => horizontal display of stimuli or vertical (depending on the aspect ratio of the screen)
	
    // compute max_w => maximum allowed image width (px) in current screen resolution
  let max_w= hor ? Math.floor(0.95 * 0.5 * win_w) : Math.floor(0.95 * win_w);	// maximum allowed image width      
		// compute max_h => maximum allowed image height (px) in current screen resolution
  let cnt = hor ? 2 : 3;   // Number of objects in the ordinate direction (y-axis)
  let max_h= Math.floor(0.95 * tot_h/cnt); // maximum allowed image height  
	let img= document.getElementById('img_cont');  // image stimulus html container					 	
  if ( stype == '1' )  // User images will be displayed (user must select image files)
	{  // rescale image to maximum allowed width and height. preserve px width/height (img scale factor)
		if (max_h < max_w)  
		{
			img.style.maxWidth= max_w + 'px'; 
			img.style.maxHeight= max_h + 'px'; 
		}
		else  {
			img.style.maxHeight= max_h + 'px';  
			img.style.maxWidth= max_w + 'px';
		} 
	}
	else  
	{ // Random alphanumeric characters will be displayed
		img.style.width=  max_w + 'px'; 
		img.style.height= max_h + 'px';  	
	}		
  let w = img.clientWidth;   // This property will round the value to an integer.
  let h = img.clientHeight;  // if img.clientWidth (w) and img.clientHeight (h) are ok, all bellow this line should be ok  
  img.style.left= Math.floor ( (win_w - w) / 2 )  + 'px' ;  // center Horizontal on screen center  
  let dy= Math.floor( (tot_h - cnt * h) / (cnt+1) ); // space between objects (for center vertically) 
	img.style.top= dy + 'px';   // center Vertically
  
  let s_canvas= document.getElementById('s_canvas');
	s_canvas.width= w;
	s_canvas.style.width= w + 'px';   	  
	s_canvas.height= h;// https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 
	s_canvas.style.height= h + 'px';  // The height property has effect only on block-level elements or on elements with absolute or fixed position.    
	s_canvas.style.left= hor ? Math.floor( win_w/4 - w/2 ) + 'px' : img.style.left; // (center Horizontal on left half of screen) or (center Horizontal on screen center)	
 	s_canvas.style.top=  Math.floor( h + 2 * dy ) + 'px';    // center Vertically  
		
	let r_canvas= document.getElementById('r_canvas');
	r_canvas.width= w;
	r_canvas.style.width= w + 'px';   	  
	r_canvas.height= h;  // The height property has effect only on block-level elements or on elements with absolute or fixed position.
	r_canvas.style.height= h + 'px'; // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 	
	r_canvas.style.left= hor ?  Math.floor( 0.75 * win_w - 0.5 * w ) + 'px' : img.style.left; // (center Horizontal on right half of screen)	or (allign left with img and s_canvas)
	r_canvas.style.top= hor ? s_canvas.style.top : Math.floor( 2 * h + 3 * dy ) + 'px';   // center vertically on screen
		  
}	


// set absolute positions with pure js for: visible  2 canvas (s_canvas + r_canvas). Model (image) is not visible
function setPosNoModelTwo()
{
	let win_w= window.innerWidth, win_h= window.innerHeight;  	
  
    // tot_h == total y height => window height - footer height	
	let tot_h= win_h - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight               
  
  let hor = win_w / tot_h  >= 1.5 ? true : false;    	// hor => horizontal display of stimuli or vertical (depending on the aspect ratio of the screen)   	  
	
    // compute max_w => maximum allowed image width (px) in current screen resolution
  let max_w= hor ? Math.floor(0.95 * 0.5 * win_w) : Math.floor(0.95 * win_w);	// maximum allowed image width      
		// compute max_h => maximum allowed image height (px) in current screen resolution
  let cnt = hor ? 1 : 2;   // Number of objects in the ordinate direction (y-axis)
  let max_h= Math.floor(0.95 * tot_h/cnt); // maximum allowed image height  
	let img= document.getElementById('img_cont');  // image stimulus html container		
	img.style.display= 'block';  // tmp		 
	if ( stype == '1' )
	{  // User images will be displayed (user must select image files)
		if (max_h < max_w)  
		{  // rescale image to maximum allowed width and height => preserve image x/y scale factor
			img.style.maxWidth= max_w + 'px'; 
			img.style.maxHeight= max_h + 'px'; 
		}
		else  {
			img.style.maxHeight= max_h + 'px';  
			img.style.maxWidth= max_w + 'px';
		}
	}
	else  
	{  // Random alphanumeric characters will be displayed
		img.style.width=  max_w + 'px'; 
		img.style.height= max_h + 'px';  	
	}		
  let w = img.clientWidth;   // This property will round the value to an integer.
  let h = img.clientHeight;  // if img.clientWidth (w) and img.clientHeight (h) are ok, all bellow this line should be ok 
  img.style.display= 'none';
  	
  let s_canvas= document.getElementById('s_canvas');
	let dy= Math.floor( (tot_h - cnt * h) / (cnt+1) ); // space between objects (for center vertically) 
	s_canvas.width= w; 
	s_canvas.style.width= w + 'px';   	  
	s_canvas.height= h;  // the height property has effect only on block-level elements or on elements with absolute or fixed position.
	s_canvas.style.height= h + 'px'; // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/      	
	s_canvas.style.left= hor ? ( Math.floor( 0.25 * win_w - 0.5 * w) + 'px' ) : ( Math.floor ( (win_w - w) / 2 )  + 'px' );     // center Horizontal on left half of screen  or  center Horizontal on screen center
	s_canvas.style.top= dy + 'px';    //  center Vertically (full y) => Math.floor( (tot_h - h) / 2 ) + 'px';
	  	  	
  let r_canvas= document.getElementById('r_canvas');
	r_canvas.width= w;
	r_canvas.style.width= w + 'px';   	  
	r_canvas.height= h;  // The height property has effect only on block-level elements or on elements with absolute or fixed position.
	r_canvas.style.height= h + 'px'; // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 
	if (hor) {
		r_canvas.style.left= Math.floor ( 0.75 * win_w  - 0.5 * w )   + 'px' ;  // center Horizontal on right half of screen 
		r_canvas.style.top= s_canvas.style.top;   // center Vertically on right half of screen == dy + 'px';
	}
	else  {
		r_canvas.style.left= s_canvas.style.left;
		r_canvas.style.top=  Math.floor( h + 2 * dy ) + 'px';   // center Vertically on screen
	}	
	
}	


// set absolute positions with pure js for: visible only one canvas (r_canvas) and model (image). s_canvas is hidden
function setPosModelOne()
{  
	let win_w= window.innerWidth, win_h= window.innerHeight;  	  
		// tot_h == total y height => window height - footer height	
	let tot_h= win_h - document.getElementById('footer').offsetHeight; // document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight               	 		  
  
  let hor = win_w / tot_h  >= 1 ? true: false;      	// hor => horizontal display of stimuli or vertical (depending on the aspect ratio of the screen) 	  
  
  let max_w= hor ? Math.floor(0.95 * 0.5 * win_w) : Math.floor(0.95 * win_w);	   // compute max_w => maximum allowed image width (px) in current screen resolution 		
  let cnt = hor ? 1 : 2;   // Number of objects in the ordinate direction (y-axis)
  let max_h= Math.floor(0.95 * tot_h/cnt); // max_h => maximum allowed image height (px) in current screen resolution  
	let img= document.getElementById('img_cont');  // image stimulus html container					  
	if ( stype == '1' )
	{  // User images will be displayed (user must select image files)
		if (max_h < max_w)    // rescale image to maximum allowed width and height => 100% fill img/canvas container.
		{
			img.style.maxWidth= max_w + 'px'; 
			img.style.maxHeight= max_h + 'px'; 
		}
		else  {
			img.style.maxHeight= max_h + 'px';  
			img.style.maxWidth= max_w + 'px';
		}
	}
  else  
	{  // Random alphanumeric characters will be displayed
		img.style.width=  max_w + 'px'; 
		img.style.height= max_h + 'px';  	
	}	
  let w = img.clientWidth;   // This property will round the value to an integer.
  let h = img.clientHeight;  // if img.clientWidth (w) and img.clientHeight (h) are ok, all bellow this line should be ok 
  if (hor) {
    img.style.left= Math.floor ( win_w/4  - w/2  )   + 'px' ;  // center Horizontal on left half of screen
  }
  else {
    img.style.left= Math.floor ( (win_w - w) / 2 )  + 'px' ;  // center Horizontal on screen center
  }  
  let dy= Math.floor( (tot_h - cnt * h) / (cnt+1) ); // space between objects (for center vertically) 
	img.style.top= dy + 'px';   // center Vertically
	
  //document.getElementById('s_canvas').style.display= 'none'; // done in startTask
  
  let r_canvas= document.getElementById('r_canvas');
	r_canvas.width= w;
	r_canvas.style.width= w + 'px';   	  
	r_canvas.height= h;  // The height property has effect only on block-level elements or on elements with absolute or fixed position.
	r_canvas.style.height= h + 'px'; 	 // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/     
	if (hor) {
		r_canvas.style.left= Math.floor ( 0.75 * win_w  - w / 2  )   + 'px' ;  // center Horizontal on right half of screen
		r_canvas.style.top= img.style.top;   // center Vertically on right half of screen => dy + 'px';
	}
	else  {
		r_canvas.style.left= img.style.left; // center Horizontal on full screen
		r_canvas.style.top= Math.floor( h + 2 * dy ) + 'px';   // center Vertically on full screen
	}		
}	


// set absolute positions with pure js for: only one visible canvas (r_canvas). Model (image) + s_canvas are not visible
function setPosNoModelOne()
{  
	let win_w= window.innerWidth, win_h= window.innerHeight;  		   	 
  let max_w= Math.floor(0.95 * win_w);	 // compute max_w => maximum allowed image width (px) in current screen resolution
	let tot_h= win_h - document.getElementById('footer').offsetHeight; // tot_h == total y height => window height - footer height.	document.getElementById('footer').clientHeight == document.getElementById('footer').offsetHeight               	 		  	   
  let max_h= Math.floor(0.95 * tot_h);   // compute max_h => maximum allowed image height (px) in current screen resolution
	let img= document.getElementById('img_cont');  // image stimulus html container		
  img.style.display= 'block'; // tmp
  if ( stype == '1' )
	{  // User images will be displayed (user must select image files)
		if (max_h < max_w)  // rescale image to maximum allowed width and height
		{
			img.style.maxWidth= max_w + 'px'; 
			img.style.maxHeight= max_h + 'px'; 
		}
		else  {
			img.style.maxHeight= max_h + 'px';  
			img.style.maxWidth= max_w + 'px';
		}
	}
	else  
	{  // Random alphanumeric characters will be displayed
		img.style.width=  max_w + 'px'; 
		img.style.height= max_h + 'px';  	
	}	
  let w = img.clientWidth;   // This property will round the value to an integer.
  let h = img.clientHeight;  // if img.clientWidth (w) and img.clientHeight (h) are ok, all bellow this line should be ok    
	img.style.display= 'none';	
     	
  let r_canvas= document.getElementById('r_canvas');
	r_canvas.width= w;
	r_canvas.style.width= w + 'px';  	 	  
	r_canvas.height= h;   // the height property has effect only on block-level elements or on elements with absolute or fixed position.
	r_canvas.style.height= h + 'px';   // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/   
	r_canvas.style.left= Math.floor ( (win_w  - w) / 2 )   + 'px' ;  // center Horizontal on full screen 
	r_canvas.style.top= Math.floor ( (tot_h  - h) / 2 )   + 'px' ;   // center Vertically on full screen 		
}	


// end of task (research)
function theEnd()  
{   

  is_task= false;   // see timer 	
	
	document.getElementById('task_section').style.display= 'none';	
	document.getElementById('s_canvas').style.display= 'none';
	document.getElementById('r_canvas').style.display= 'none';
	document.getElementById('m_canvas').style.display= 'none';  //document.getElementById('footer').style.display= 'none';
  document.getElementById('end').style.display= 'block';   
    
  let cor = 0, rows = [], moves_ok= [], rot_ok= [], completed = [], desc, sum_moves=0, sum_rot=0;   
	// compute points (score) for task. For each item: if the item was successfully completed = 1 point. Else = 0 points. 
	for (let i=0; i<item_cnt; i++)  
	{
		completed[i]= (itemDone(i) === true) ? 1 : 0; 
		desc= completed[i] === 1 ? 'Yes' : 'No';  
    sum_moves+= moves[i]; // sum of all moves during task
    sum_rot+= rot_cnt[i];		// sum of all user done rotation during task
    if (completed[i] === 1)  
    { 
      cor++;  // cor => number of completed items => number of points in task
      moves_ok.push(moves[i]);   // moves for this completed item (only for completed items)
      if (rotation) rot_ok.push(rot_cnt[i]);   // rotation done (user dblCick) by user (only for completed items)
    }       
    if (rotation)  {
      rows[i]= [ parseInt(i+1,10), stim[i], desc, completed[i], moves[i], rot_cnt[i] ];  // => [ 'Item', 'Stimulus' , 'Completed', 'Points', 'Moves', 'Rotation-Moves' ] 			 
    }
    else  {
      rows[i]= [ parseInt(i+1,10), stim[i], desc, completed[i], moves[i] ];  // => [ 'Item', 'Stimulus' , 'Completed', 'Points', 'Moves' ]        
    }		
	}		
	if (rotation)  {
    rows.push( [ 'Sum', '', '', cor, sum_moves, sum_rot] );	// add last summary row:	Sum for: Points, Moves, Rotation-Moves
 	}
  else  {
    rows.push( [ 'Sum', '', '', cor, sum_moves, '' ] );	// add last summary row: Only Sum for Points and Moves
  }
  document.getElementById('result_wrapper').style.display= 'block';  
	if (rotation)  {
		createTableArray(document.getElementById('table1'), rows,  ['Item', 'Stimulus' , 'Completed', 'Points', 'Moves', 'Rotation-moves' ]); 
	}
	else {
		createTableArray(document.getElementById('table1'), rows,  ['Item', 'Stimulus' , 'Completed', 'Points', 'Moves' ]);  
	}
	document.getElementById('table1').style.display= 'block';  	
	
	// calculate decriptive statistics for the moves [and rotation] ONLY on completed items => average and sample standard deviation..	
  let n = moves_ok.length; 
  let m = n > 0 ? mean(moves_ok).toFixed(3) : '';  // mean for number moves. Only for completed items
	let sd = n > 1 ? stDev(moves_ok, m).toFixed(3) : ''; // standard deviation for number of moves. Only for completed items   
	n= rot_ok.length; 
  let m_rot =  n > 0 ? mean(rot_ok).toFixed(3) : '';  // mean for number of rotation done (user dblCick). Only for completed items
	let sd_rot = n > 1 ? stDev(rot_ok, m).toFixed(3) : ''; // standard deviation for for number of rotation done (user dblCick). Only for completed items 
	
  let tmp=[];  //'Cannot be calculated'
	tmp[0]= ['Moves', m, sd, cor], tmp[1]= ['Rotation', m_rot, sd_rot, cor];		
	createTableArray(document.getElementById('table2'), tmp, ['Variable', 'Average', 'Sample Standard Deviation', 'N']);   
	
  // "chart" => for points - repaint "chart" 
  let percent= (100*cor/item_cnt).toFixed(3), not_done= (100-percent).toFixed(3);	
	document.getElementById('chart_title').textContent= 'Percentage of completed items (green)= ' + percent + '% and not completed items (red)= ' + not_done + '%';
	document.getElementById('chart_cor').setAttribute('style', 'width:' + percent + '%');    // !important
	document.getElementById('chart_err').setAttribute('style', 'width:' + not_done + '%');
  
  // update (draw)  "chart2" => for moves. Only for completed items     
  let eff= '', redundancy= '';
  if ( cor > 0 )  {
		document.getElementById('chart2_wrapper').style.display= 'block';
		eff= ( 100 * (raster * raster) / m   ).toFixed(3);
    if (eff > 100) eff= 100;   // this is possible if mean for number moves is lower then raster*raster
    redundancy= (100- eff).toFixed(3); // percentage of efficiency, percentage of redundant moves	
		document.getElementById('chart2_title').textContent= 'Moves only for completed items: Percentage of efficacy (green)= ' + eff + '% and Percentage of redundant moves (blue)= ' + redundancy + '%';
		document.getElementById('chart2_eff').setAttribute('style', 'width:' + eff + '%');   
		document.getElementById('chart2_red').setAttribute('style', 'width:' + redundancy + '%');
	}
	else {
		document.getElementById('chart2_wrapper').style.display= 'none';			
	}
 	
	try 
  {
    if (store) {
			setStorage(cor, eff, redundancy, m, sd, m_rot, sd_rot); 
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }    	 
  
  document.body.style.backgroundColor= '#ffffff';  // restore white color (at the end - wait for other process)
}


// return arithmetic mean (mean, average) of array (arr) 
function mean(arr)  
{	
	let sum=0;
	for (let i in arr)  {
		sum+= parseFloat(arr[i]);
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation
// arr = array argument; mean = average (mean)
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
	  sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 point  => number of points == number of completed items (user get 1 point for each completed item)
   eff => percentage of efficiency
   red => percentage of redundant moves
   m_mov => mean for number moves. Only for completed items
   sd_mov => standard deviation for number of moves. Only for completed items   
   m_rot => mean for number of rotation done (user dblCick). Only for completed items and rotation enabled
   sd_rot => standard deviation for for number of rotation done (user dblCick). Only for completed items and rotation enabled 	
  */  
function setStorage(point, eff, red, m_mov, sd_mov, m_rot, sd_rot)   
{  
  let oldItems = JSON.parse(localStorage.getItem("wm15-v1.0")) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		   
      'points' : point, 	
			'efficiency' : eff,
      'redundancy': red,
      'M-moves' : m_mov, 
			'SD-moves' : sd_mov, 
      'M-rot' : m_rot, 
			'SD-rot' : sd_rot,      
    	'items' : item_cnt,  			
			'stimulus-type' : stype,
			'raster' : raster,
      'model' : model ? 1 : 0,
      'two' : two ? 1 : 0,		
			'rotation' : rotation ? 1 : 0,
      'zoom' : (rotation && zoom_out) ? 1 : 0,
      'angle' : rotation ? ang : 0, 		
      'demo' : demo  ? 1: 0,
      'timer-time' : timer_enabled ? document.getElementById("timer").value : '',     
		 	'is-back' : (is_footer && is_back) ? 1: 0,
		  'is-forward' : (is_footer && is_forward) ? 1: 0,	
     	'alfa-type': stype === '0' ? alfa_type : '', 
		  'is-grid' : is_border ? 1 : 0   			
  };

  oldItems.push(newItem);

  localStorage.setItem("wm15-v1.0", JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
  	
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement("table");  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear wm15 web browser storage data
function clrStorage()  
{  
  let cnt= localStorage.getItem('wm15-v1.0')  ?  JSON.parse(localStorage.getItem('wm15-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no wm15 data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' wm15 results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved wm15 results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('wm15-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data  ';
 }  
}


// clear the canvas object. argument e == html canvas object/element
function clear(e)  {	 
  let ctx= e.getContext('2d');
	ctx.clearRect(0, 0, e.clientWidth, e.clientHeight);
}	


// draws a raster of rectangles on canvases
function gridRect()
{
  let r_canvas= document.getElementById('r_canvas'), ctx_r= r_canvas.getContext('2d'), 
	    w = r_canvas.width, h = r_canvas.height, dx = w / raster, dy = h / raster;
	//ctx_r.strokeStyle= '#000000'; ctx_r.lineWidth=1;	
	for (let row = 0; row <= raster; row++ )  {	    	
		ctx_r.beginPath();
    ctx_r.moveTo(0, row * dy);		
    ctx_r.lineTo(w, row * dy);		
		ctx_r.stroke();
	}	  
	for (let col = 0; col <= raster; col++)  {			
		ctx_r.beginPath();
    ctx_r.moveTo(col*dx, 0);		
    ctx_r.lineTo(col*dx, h);		
		ctx_r.stroke();
	}
	
	if (two)  {
		let s_canvas= document.getElementById('s_canvas'), ctx_s= s_canvas.getContext('2d');
	  //ctx_s.strokeStyle= '#000000'; ctx_s.lineWidth=1;
		for (let row = 0; row <= raster; row++ )  {	    	
			ctx_s.beginPath();
			ctx_s.moveTo(0, row * dy);		
			ctx_s.lineTo(w, row * dy);		
			ctx_s.stroke();
		}	  
		for (let col = 0; col <= raster; col++)  {			
			ctx_s.beginPath();
			ctx_s.moveTo(col*dx, 0);		
			ctx_s.lineTo(col*dx, h);		
			ctx_s.stroke();
		}
	}		
}


/* EventListener for input type="file" onchange. see html: document.getElementById("add_file")
   When the user selects image files
   https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsDataURL */
/* EventListener for input type="file" onchange. see html: document.getElementById("add_file")
   When the user selects image files
   https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsDataURL */
function createImages() 
{
  let preview = document.querySelector('#preview');
  let files = document.querySelector('input[type=file]').files;
  let cnt= document.getElementById('preview').getElementsByTagName('img').length;  
	function readAndPreview(file) 
	{    
    if ( /\.(jpe?g|png|gif|bmp)$/i.test(file.name) ) {  // Make sure `file.name` matches our extensions criteria
      let reader = new FileReader();
      reader.addEventListener("load", function () {
        let image = new Image();              
        image.src = this.result;        
				image.setAttribute('class', 'flex-child');   //image.height = 100;
				image.title = 'item ' + parseInt(cnt+1) + '=' + file.name;
				//image.setAttribute('data-title', 'item ' + parseInt(cnt+1)); 
				// drag and drop  
				image.id= 'img' + parseInt(cnt+1); 
				image.setAttribute("draggable", "true"); 
				image.setAttribute("ondragstart", "onDragStart(event)");   				
				image.setAttribute("ondragover", "allowDrop(event)");
				image.setAttribute("ondrop", "onDrop(event)");
				//image.setAttribute("ondblclick", "delItem(this)");     
        
				preview.appendChild(image);
				cnt++;
				if (cnt>0) document.getElementById('img_info').style.display= 'block';
				document.getElementById('item_cnt').value= cnt; 
				itemsValid();   //document.getElementById('item_cnt_msg').textContent= '';
        refMsg();
      }, false);
      reader.readAsDataURL(file); 
    }
  }
  if (files) {
    [].forEach.call(files, readAndPreview);
  }
  dirty=true;  
	//document.getElementById('item_cnt').value= files.length;       	
}


// ondragstart event for stimulus type= images
function onDragStart(ev)  {   
  ev.dataTransfer.setData('text', ev.target.id); 
}


// remove image file from container (stimulus type: images). see html source
function delItem(e)
{ 
	e.preventDefault(); 
  let my_id= e.dataTransfer.getData('text');    // node1.id
  let node1= document.getElementById(my_id);    // node1= DRAG node pointer             		
	node1.parentNode.removeChild(node1);   // delete ex node2
	toogleInfo();
  dirty=true;  
  refMsg();	
}


// ondrop() event for stimulus type= images
// e => html image element
function onDrop(e)  
{  
	e.preventDefault(); 
  let my_id = e.dataTransfer.getData('text');    // node1.id
  let node1= document.getElementById(my_id);    // node1= DRAG node pointer           
	let node2= e.target;   //  node2= DROP node pointer: above  <ol> or <li> ??  => [object HTMLOListElement] or [object HTMLLIElement] => this is drop node      
	let parNode= document.getElementById('preview');  
  if ( node2.tagName === "IMG" )  {
    if ( node2 === parNode.lastChild ) {
      insertAfter(node1, node2);  // append image to end of list
    } 
    else {
      node2.parentNode.insertBefore(node1, node2);   // move (insert) node1 before node2 => let insertedNode = parentNode.insertBefore(newNode, referenceNode)
    }	    
  }
  toogleInfo();
}


// refNode is the node you want to put newNode after. newNode => new node. insert new node after refNode
function insertAfter(newNode, refNode) {
  refNode.parentNode.insertBefore(newNode, refNode.nextSibling);
}


// auto set item_cnt in task based on current number of images stimulus
// set visibility for small info div: "Drop image in this line here to delete" and "Use drag and drop to move files (stimulus)"
function toogleInfo()  
{
	let img_cnt= document.getElementById('preview').getElementsByTagName('img').length;  // current number of images stimulus
	document.getElementById('item_cnt').value= img_cnt;
  document.getElementById('img_info').style.display= img_cnt > 0 ? 'block' : 'none';
	itemsValid();
}	


// creates a set of rectangles within an r_canvas bounds. r_canvas is always visible
// warning: each time on <body onresize="pageResize()"> is executed it will call drawStim() => loadImage() =>  setPositions().
// result: r_canvas.width and r_canvas.height will be changed. (they depends on currently loaded img.clientWidth, img.clientHeight)   
function setRects()
{
	let rects = [], r_canvas= document.getElementById('r_canvas'), dx = r_canvas.width / raster, dy = r_canvas.height / raster;
	for (let row = 0; row < raster; row++ )
	{	    	
		for (let col = 0; col < raster; col++)
		{													
			let x = col * dx;
			let y = row * dy;  
			let	w = dx;
			let	h = dy;
			rects.push({
				x0:x,         
				y0:y, 
				w: dx,
				h: dy   
			});			     				
		}		
	}	  
  return rects;  // return an array of objects that define different rectangles
}


// init parts with raster of rectangles using conditional random puzzle generator. 
// this function is executed only once within each item
function initParts()
{	   				
	let last= raster*raster-1, j, sel = {}, not_ok;  		
  for (let i=0; i <= last; i++)			//  i => rect (image part) counter,  j= random image part, sel => previously selected image rects 
	{	 
    if (!demo)
		{
			do  {
				j = getRndInt(0, last);		// choose image random rect region			
				not_ok = (typeof sel[j] != "undefined") ||  ( j == i && i < last);			
			}	while ( not_ok );
			sel[j]= 1; 						  	
    }
		else  {
			j= i;
		}		
		if (demo)  {
			r_parts[item][i][0]= j;   // correct position
      //r_parts[item][i][1]= 0;   // current angle = 0 (no rotation) => already set in setPar()					
		}
		else  
		{
			if (two)  {
				s_parts[item][i][0]= j;   // correct position
				//s_parts[item][i][1]= 0;   // current angle = 0 (no rotation) => already set in setPar()		      
			}
			else  {
				r_parts[item][i][0]= j;   // correct position
				//r_parts[item][i][1]= 0;   // current angle = 0 (no rotation) => already set in setPar()				
			}
		}		
	}   
} 
   

// init rotated parts with raster of rectangles using conditional random puzzle generator. 
// this function is executed only once within each item
function initRotatedParts()
{		 				
	let last= raster*raster-1, up= angles.length, i= 0, j, sel = {}, not_ok, ind;  //  i => rects (image parts) counter,  j= random image part, sel => previously selected image rects 			
	for (let i=0; i <= last; i++)
	{	    		     
		if (!demo)
		{
			do  {
				j = getRndInt(0, last);		// choose image random rect region			
				not_ok = (typeof sel[j] != "undefined") ||  ( j == i && i < last);			
			}	while ( not_ok );
			sel[j]= 1;  	
			ind= Math.floor(Math.random()*up);   // getRndInt(0,up-1); 
		}
		else  {
			j= i;	
		}	
		if (demo)  {
			r_parts[item][i][0]= j;   // correct position
      //r_parts[item][i][1]= 0;   // angle (no rotation). already set in setPar() 					
		}
		else  
		{
			if (two)  {
				s_parts[item][i][0]= j;            // correct position
				s_parts[item][i][1]= angles[ind];  // rotation angle in degrees. no rotation for demo= true	        
			}
			else  {
				r_parts[item][i][0]= j;            // correct position
				r_parts[item][i][1]= angles[ind];  // rotation angle in degrees. no rotation for demo= true			
			}
		}		
	} 
} 


/* show current puzzle positions (when moving between items)
	 void ctx.drawImage(image, dx, dy);
	 void ctx.drawImage(image, dx, dy, dWidth, dHeight);
	 void ctx.drawImage(image, sx, sy, sWidth, sHeight, dx, dy, dWidth, dHeight);  */
function showParts()
{ 
	let rects = setRects(); 	   
	let img= document.getElementById('img_cont');  // image stimulus html container 	
  let dx= img.naturalWidth / raster;   
	let dy= img.naturalHeight / raster;  
	let len= raster*raster, pos, row, col;   //  total image parts count, part position, row count, col count 
	if (two)
	{
		let s_canvas= document.getElementById('s_canvas');
		clear(s_canvas);
		let ctx_s= s_canvas.getContext('2d');
		for (let i=0; i < len; i++)
		{	    		     
			pos= s_parts[item][i][0];
			if (pos != -1)
			{			 
				row = Math.floor (pos / raster);
				col = Math.floor (pos % raster);  // or => col= pos - row * raster;  // pos = row * raster + col;               		
				ctx_s.drawImage(img, col*dx, row*dy, dx, dy,   rects[i].x0, rects[i].y0, rects[i].w, rects[i].h);	 			
			}		
		}  
	} 
  let r_canvas= document.getElementById('r_canvas');
	clear(r_canvas);
	let ctx_r= r_canvas.getContext('2d');  	
	for (let i=0; i < len; i++)
	{	    		     
    pos= r_parts[item][i][0];
		if (pos != -1)
		{			 
			row = Math.floor (pos / raster);
			col = Math.floor (pos % raster);  // or => col= pos - row * raster;  // pos = row * raster + col;      			         		
			ctx_r.drawImage(img, col*dx, row*dy, dx, dy,   rects[i].x0, rects[i].y0, rects[i].w, rects[i].h);      		
		}		
	} 		  
}	


// show current rotatated puzzle positions (when moving between items)
function showRotatedParts()
{ 
	let rects = setRects(); 	   
	let img= document.getElementById('img_cont');  // image stimulus html container 	
  let dx= img.naturalWidth / raster, dy= img.naturalHeight / raster;  
	let len= raster*raster, pos, row, col, x, y, ctx;   //  total image parts count, part position, row count, col count 
	if (two)
	{
		let s_canvas= document.getElementById('s_canvas');
		clear(s_canvas);
		ctx= s_canvas.getContext('2d');
		for (let i=0; i < len; i++)
		{	    		     
			pos= s_parts[item][i][0];
			if (pos != -1)
			{			 
				row = Math.floor (pos / raster);
				col = Math.floor (pos % raster);                 									
        x= rects[i].x0 + rects[i].w/2;  // x center of destination rectangle 
				y= rects[i].y0 + rects[i].h/2;  // y center of destination rectangle 								
				ctx.save();  				
				ctx.setTransform(1, 0, 0, 1, x, y);   //ctx.translate(x,y);					
				ctx.rotate(s_parts[item][i][1] * Math.PI/180);  // s_parts[item][i][1] is always in degrees
				ctx.drawImage(img, col*dx, row*dy, dx, dy,   rects[i].x0-x, rects[i].y0-y, rects[i].w, rects[i].h);   
				ctx.restore();	
			}		
		}
    ctx.setTransform(1, 0, 0, 1, 0, 0); // Reset transformation matrix to the identity matrix 		
	} 
	
	let r_canvas= document.getElementById('r_canvas');
	clear(r_canvas);
	ctx= r_canvas.getContext('2d');
	for (let i=0; i < len; i++)
	{	    		     
		pos= r_parts[item][i][0];
		if (pos != -1)
		{			 
			row = Math.floor (pos / raster);
			col = Math.floor (pos % raster);                 								
			x= rects[i].x0 + rects[i].w/2;  // x center of destination rectangle 
			y= rects[i].y0 + rects[i].h/2;  // y center of destination rectangle 								
			ctx.save();  				
			ctx.setTransform(1, 0, 0, 1, x, y);   //ctx.translate(x,y);					
			ctx.rotate(r_parts[item][i][1] * Math.PI/180);  // r_parts[item][i][1] is always in degrees
			ctx.drawImage(img, col*dx, row*dy, dx, dy,   rects[i].x0-x, rects[i].y0-y, rects[i].w, rects[i].h);   
			ctx.restore();	
		}		
	}  
	
	ctx.setTransform(1, 0, 0, 1, 0, 0); // Reset transformation matrix to the identity matrix 
}	


// set r_canvas bounds for output of drawText()
function setTextBounds()  
{
  let r_canvas= document.getElementById('r_canvas');
  let ctx= r_canvas.getContext('2d');  
	let imageData= ctx.getImageData(0, 0, r_canvas.width, r_canvas.height);		

	let res= getBlackRect();  // res => ( [0] = first row = y, [1] = first column = x, [2] = last row = y, [3]= last column= x) with black pixels   	

	let off= 10; // offset around text (px)
	
  let bounds= {left:res[1]-off, top: res[0]-off, width:res[3]-res[1]+2*off, height: res[2]-res[0]+2*off};	//bounds ={left:x0, top: y0, width: w, height: h};		
  
  return bounds;
}


/* use only for black pixels search on window.r_canvas object (html r_canvas element)
   return rect bounds ( [0] = first row = y, [1] = first column = x, [2] = last row = y, [3]= last column= x) with black pixels 
   return [left, top, right, bottom] corners for black pixels */
function getBlackRect()
{
  let r_canvas= document.getElementById('r_canvas');
  let ctx= r_canvas.getContext('2d');	
	let pixels= ctx.getImageData(0, 0, r_canvas.width, r_canvas.height).data;
  let cnt= 0, row= 0, col= 0; 
	let res=[-1, -1, -1, -1];  // left, top, right, bottom corners
  let len= pixels.length;
	let w= r_canvas.width;	
	for (let i = 0; i != len; i += 4) 
	{		
		if ( res[0] == -1 && pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			res[0]= row;   // find first row with minimum one black pixel   
		}
    else if ( pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			res[2]= row;   // update (this) last row with minimum one black pixel   
		}			
		if ( (res[1] == -1 || col < res[1]) && pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			res[1]= col;  // find first column with minimum one black pixel  
		}
		else if ( col > res[3] && pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			res[3]= col;  // find last column with minimum one black pixel  
		}	
	  cnt++;    // total pixel counter
		col++;    // column counter
		if ( cnt % w == 0 ) {   // enter new row
			row++;	
			col=0;
		}	
	}
	row--;			
	return res;
}	


// tmp copy a part of r_canvas to image
function canvasToImage(bounds)
{  
  let r_canvas= document.getElementById('r_canvas');
  let ctx= r_canvas.getContext('2d');
	let imgData = ctx.getImageData(bounds.left, bounds.top, bounds.width, bounds.height);  // get alfanumeric text content
	ctx.clearRect(0, 0, r_canvas.clientWidth, r_canvas.clientHeight);  // make invisible alfanumeric stimulus early as possible	
	document.getElementById('img_cont').src = getImageURL(imgData, bounds.width, bounds.height);    
}	


// copy a part of r_canvas to image
// https://stackoverflow.com/questions/45234492/copy-a-part-of-r_canvas-to-image
function getImageURL(imgData, width, height) {
   let mCanvas = document.getElementById('m_canvas')
   let ctx = mCanvas.getContext('2d');
   mCanvas.width = width;
   mCanvas.height = height;    //ctx.fillStyle = "#ffffff";
	 ctx.clearRect(0, 0, width, height);   
   ctx.putImageData(imgData, 0, 0);
   return mCanvas.toDataURL();  //image URL
}


/* mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
   e => html canvas => document.getElementById('s_canvas') or document.getElementById('r_canvas') */
function showCanvas(e)
{  
  let m_canvas = document.getElementById('m_canvas');
  let s_canvas = e.target || e.srcElement;   // document.getElementById('s_canvas') or document.getElementById('r_canvas')   
	
	let r = getRectFromXY(s_canvas,  e.offsetX, e.offsetY);
  
	if ( typeof(r) === 'undefined' || typeof(r[2]) === 'undefined' || r[2] === -1 )  {
    m_canvas.style.display= 'none';	
		return;  // return 'undefined' value
	}		
  let bounds= s_canvas.getBoundingClientRect();
  let offsetX = bounds.left, offsetY = bounds.top;	
	let row = parseInt( r[0],10 ), col = parseInt( r[1],10 );      // part (rect) ordinal position within canvas         
  let dx = s_canvas.width / raster, dy = s_canvas.height / raster; 
     
  // set absolute positions (left, top, width, height) for m_canvas object
	m_canvas.style.left= parseInt(col*dx + offsetX, 10)  + 'px';  // top and left properties require a unit of measure, not just a number
  m_canvas.style.top= parseInt(row * dy + offsetY) + 'px';   
  /* the bottom 2 lines of code position the rectangle (left,top) exactly where the user clicked
  m_canvas.style.left= e.clientX + 'px';  // top and left properties require a unit of measure, not just a number
  m_canvas.style.top= e.clientY + 'px';  */ 
  m_canvas.width= dx;
  m_canvas.height= dy;    
  m_canvas.style.display= 'block';	
	
	r.push(s_canvas.id);  // add source canvas.id
	
	m_canvas.setAttribute('data-r', r);   //  xxx  'data-r' = [0 => row; 1 = col; 2 = correct position; 3= rotation angle; 4 = canvas.id]   xxx 
	
  clear(m_canvas);  // clear content of m_canvas
 	let pos= row * raster + col;  // read rotation angle from s_canvas row and col (not from image)
 	let angle =  (s_canvas.id == 's_canvas') ? parseInt(s_parts[item][pos][1],10) : parseInt(r_parts[item][pos][1],10);  // angle in degrees 
	
	angle= angle * Math.PI/180;  // compute angle in radians (because s_parts and r_parts are always in degrees)
	
	pos= 	parseInt(r[2],10);  // pos is part ordinal position in image (not in canvas)  	
  row = Math.floor (pos / raster),  col = Math.floor (pos % raster);  // update row and column (from image not from canvas) 
	
	let img= document.getElementById('img_cont');
	dx= img.naturalWidth / raster,  dy= img.naturalHeight / raster;  
	let ctx= m_canvas.getContext('2d');
  let	w= m_canvas.width, h= m_canvas.height;									               
	let x= w/2, y= h/2, scale= 1;   // x and y center of destination rectangle, scale factor  								
  if (rotation && zoom_out)  {
    let d= Math.sqrt ( Math.pow(dx, 2) + Math.pow(dy, 2) );    // d = hypotenuse = the length of the diagonal by the Pythagorean theorem.  Math.pow(base, exponent)		
    scale= Math.min(dx, dy) / d;  // axis scale factor =>  min(width,height)/ hypotenuze length	
  }	
	
	ctx.save();  				
	ctx.setTransform(scale, 0, 0, scale, x, y);   //ctx.translate(x,y);					
	ctx.rotate(angle);  // 
	ctx.drawImage(img, col*dx, row*dy, dx, dy,   -x, -y, w, h);     // ex: ctx.drawImage(s_canvas, col*dx, row * dy, dx, dy,     0, 0, dx, dy);
	
	ctx.restore();			
	ctx.setTransform(1, 0, 0, 1, 0, 0); // Reset transformation matrix to the identity matrix 
		      
}


/* Get the rectangle (parts) row, col and ordinal position in html canvas element from x and y mouse position
   Arguments: (canvas = html canvas element => document.getElementById('s_canvas') or document.getElementById('r_canvas') 
   The offsetX (offsetY) property returns the x (y) -coordinate of the mouse pointer, relative to the target element.  */
function getRectFromXY(canvas, offsetX, offsetY)
{	
  let dx =  canvas.width / raster, dy = canvas.height / raster;  
  let row = Math.floor ( offsetY / dy );  // The offsetX (offsetY) property returns the x (y) -coordinate of the mouse pointer, relative to the target element. 
  let col = Math.floor ( offsetX / dx );  // The clientX (clientY) property returns the horizontal (vertical) coordinate (according to the client area) of the mouse pointer when a mouse event was triggered.
  if (row > raster -1 || row < 0 || col > raster -1 || col < 0) {
    return;  // return 'undefined' value	=> abort all     
  }
  let pos = row * raster + col;   // rectangle position within html canvas object  	
	if ( pos >= raster * raster || pos < 0 || typeof ( s_parts[item][pos][0] ) === 'undefined' || typeof ( r_parts[item][pos][0] ) === 'undefined'
    ||  typeof ( s_parts[item][pos][1] ) === 'undefined' || typeof ( r_parts[item][pos][1] ) === 'undefined'	)  {
    return; // return 'undefined' value	=> abort all  
  }
	let cor= canvas.id === 's_canvas' ?  s_parts[item][pos][0] : r_parts[item][pos][0] ;  // correct position of this part (rectangle)
	let angle= canvas.id === 's_canvas' ?  s_parts[item][pos][1] : r_parts[item][pos][1] ;  // rotation angle of this part (0 or Math.PI)
	let res= [row, col, cor, angle];  // [0 => row; 1 = col; 2 = correct position; 3= rotation angle]    
  return res;	
}	


// ondragstart event for s_canvas and r_canvas
function dragStart(e)  {   
     //  xxxxx  'data-r' = [0 => row; 1 = col; 2 = correct position; 3= rotation angle; 4 = canvas.id]   xxxxx 
	e.dataTransfer.setData('text', document.getElementById('m_canvas').getAttribute('data-r') );
    //  xxxxx ' new e.dataTransfer.data = [0 => row; 1 = col; 2 = correct position; 3= rotation angle; 4 = canvas.id]  xxxxx
}


// ondragover() event for s_canvas and r_canvas
// the ondragover event specifies wher the dragged data can be dropped
function allowDrop(e)  {
  e.preventDefault();
}


// ondrop() event for s_canvas and r_canvas
function drop(e)  
{  
  e.preventDefault();    
  let d_canvas = e.target || e.srcElement;  // document.getElementById('r_canvas') or document.getElementById('s_canvas');      
  let d_data = getRectFromXY(d_canvas, e.offsetX, e.offsetY);    	
  if ( d_canvas.id != 's_canvas' && d_canvas.id != 'r_canvas' || typeof(d_data) === 'undefined' || typeof(d_data[3]) === 'undefined' )  {
		return;
	}	
  let row = parseInt(d_data[0], 10), col = parseInt(d_data[1], 10), d_pos = row * raster + col;   // row, col, position within DESTINATION html canvas object
	 
	clear(m_canvas);  // clear m_canvas content
	m_canvas.style.display= 'none';	
	  
	let s_data= e.dataTransfer.getData('text').split(',');  //  data from start or SOURCE rect. 
	for (let i=0; i<4; i++)  {  
		s_data[i]= parseFloat(s_data[i]);    //  xxxxx s_data = [0 => row; 1 = col; 2 = correct position; 3= rotation angle; 4 => canvas.id]  xxxxx
	}  	
	let s_canvas= document.getElementById(s_data[4]);  // source (start) canvas.id  
	let s_pos= s_data[0] * raster + s_data[1];      // s_pos = row * raster + col; 	
	
	// set window.parts array (rects positions and angles rotation)
  // a) set parts positions and angles  for source rect:	 
	if (s_canvas.id == 's_canvas')  {   
	  s_parts[item][s_pos][0]= d_data[2];   // update source part correct position
    s_parts[item][s_pos][1]= d_data[3];		// update source part rotation angle
	}
	else  {
		r_parts[item][s_pos][0]= d_data[2];   // update source part correct position
		r_parts[item][s_pos][1]= d_data[3];		// update source rotation angle
	}
	// b) set parts positions and angles for destination rect: 
	if (d_canvas.id == 'r_canvas')  {   // user is drop part (rect) into result r_canvas 
		r_parts[item][d_pos][0]= s_data[2];   // update destination part correct position
    r_parts[item][d_pos][1]= s_data[3]; 	// update destination part rotation angle	
  }
	else  {  // user is drop part (rect) into source s_canvas 
		s_parts[item][d_pos][0]= s_data[2];   // update destination part correct position
		s_parts[item][d_pos][1]= s_data[3];   // update destination part rotation angle
	}
  
  moves[item]++;
   
	if (rotation)  {
		showRotatedParts();  
	}
	else  {
		showParts();
	}		 
	 	   	
	if (is_border) {
		gridRect();
	}
	
	if (itemDone(item))  {
    document.body.style.backgroundColor= '#2E8B57';  // seagreen
    delay_cnt= 0;
    go_forward= true;
  }	 
}


// i == ordinal number of item in task. return true if the i-item (image) is successfully solved.
function itemDone(i)
{
	for (let pos in r_parts[i])  {
		if (r_parts[i][pos][0] != pos || r_parts[i][pos][1] != 0)  {
			return false;
		}		
	}
	return true;
}


// ondblclick event occurs when the user double-clicks on an m_canvas
function dblClick(e)  
{  
  if (!rotation)  {
		return;
	}
	let m_canvas= document.getElementById('m_canvas'), angle;
	let r = m_canvas.getAttribute('data-r').split(',');    //  xxxxx  'data-r' = [0 => row; 1 = col; 2 = correct position; 3= rotation angle; 4 = canvas.id]   xxxxx 
	
	let row= parseInt(r[0],10), col= parseInt(r[1],10), s_canvas= document.getElementById(r[4]);  // s_canvas.id = 's_canvas' or 'r_canvas'
	let i = row * raster + col;  // source canvas (s_canvas or r_canvas) rect ordinal position (this is not correct position)
  
	// 1. update rotation angle 
	if (s_canvas.id == 's_canvas')  {		// s_canvas.id = 's_canvas' or 'r_canvas'
		s_parts[item][i][1]+= ang;
		if ( s_parts[item][i][1] >= 360 ) {
			s_parts[item][i][1]= 0;
		}
		angle= s_parts[item][i][1];  // angle in degrees		
	}
	else  {
		r_parts[item][i][1]+= ang;
		if ( r_parts[item][i][1] >= 360 ) {
			r_parts[item][i][1]= 0;
		}			
		angle= r_parts[item][i][1];  // angle in degrees		
	}
	
	r[3]= angle;  // update new angle
	document.getElementById('m_canvas').setAttribute('data-r', r);  // update rect data for possible latter use "drag and drop" operation
	rot_cnt[item]++;
	
	showRotatedParts();   // 2. repaint two canvas again
		
	// update new rotated image part displayed on m_canvas 
	clear(m_canvas);  // clear content of m_canvas
 	let pos= parseInt(r[2],10);  // pos is part ordinal position in image (not in canvas)  	
	row = Math.floor (pos / raster),  col = Math.floor (pos % raster);  // update row and column (from image not from canvas) 
	let img= document.getElementById('img_cont'),	dx= img.naturalWidth / raster,  dy= img.naturalHeight / raster;  
	let ctx= m_canvas.getContext('2d'), w= m_canvas.width, h= m_canvas.height;									               								
	angle= angle * Math.PI/180;  // compute angle in radians (because s_parts[item][i][1] and r_parts[item][i][1] are always in degrees)
  let x= w/2, y= h/2, scale= 1;   // x and y center of destination rectangle, scale factor 	  	 				  
  if (zoom_out)  {
    let d= Math.sqrt ( Math.pow(dx, 2) + Math.pow(dy, 2) );    // d = hypotenuse = the length of the diagonal by the Pythagorean theorem.  Math.pow(base, exponent)		
    scale= Math.min(dx, dy) / d;  // axis scale factor =>  min(width,height)/ hypotenuze length	
  }
  ctx.save(); 
  ctx.setTransform(scale, 0, 0, scale, x, y);   
	ctx.rotate(angle);   
	ctx.drawImage(img, col*dx, row*dy, dx, dy,   -x, -y, w, h);   
	ctx.restore();			
	ctx.setTransform(1, 0, 0, 1, 0, 0); // Reset transformation matrix to the identity matrix 
	 	
  if (is_border) {
		gridRect();
	}
	 
  if (itemDone(item))  {
    document.body.style.backgroundColor= '#2E8B57';  // seagreen   
    delay_cnt= 0;
    go_forward= true;
  }
 
}
